#!/usr/bin/env pypy

###############################################################################
# Mjöllnir -- a simulator of the magnetic properties of polynuclear complexes
# Copyright (c) 2009--2020 Anton S. Lytvynenko, Elena A. Mikhalyova
#
# The authors thanks to Dr. V. A. Bogaenko and to Sergei Litvinenko for
# technical help and valuable discussion.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

import sys,os
import xml.dom.minidom as xmlparser
import math

from mj_matrix_operators import fl_equal, z, z_square, z1z2, plus, minus, plusminus, minusplus, const_by_matrix, filter_zeros, merge_sum, merge_by_coef
from mj_const import *
from mj_call_eigensolver import mj_eigensolver_caller
from mj_config import *

def pl_add(param_list, el):
    '''List building with duplicates removal (a bit dummy but will do for our purposes).'''
    if not el in param_list:
        param_list.append(el)

def check_pos_one_half(s):
    ''' Values of quantum numbers S, L or J have to be integer or half-integer positive numbers.
        This function checks, if they are, and returns the number if yes. Argument s is a string.'''
    a = float(s)
    if a < 0:
        raise Exception('Negative data')
    if not fl_equal(2*a,round(2*a)):
        raise Exception('Non one-half data')
    return a

def projections_list(moment):
    '''Takes S,L or J moment value and generates a list of M_S, M_L or M_J,
    E.g.: S = +5/2 -=> +5/2,+3/2,+1/2,-1/2,-3/2,-5/2.'''
    proj_list = []
    j = moment
    while j >= -moment or fl_equal(j,-moment):
        proj_list.append(j)
        j-=1
    return proj_list

def list_append2copy(l, a):
    '''An auxiliary function that takes a list l, copies it into a new list, appends a and 
    returns the result'''
    l_copy = l[:] # Copying a list with creation of a new one, "l_copy = l" just copies reference
    l_copy.append(a)
    return l_copy

class mj_fragment_matrix:
    'A distinct fragment of Mjollir model with xT calculation based on matrices diagonalization'
    def __init__(self, frag_num):
        self.frag_num = frag_num
        self.frag_type = "matrix"
        self.zeeman_symmetry = "isotropic" # Initial value, will be filled by preparse_atoms_list()
        self.total_l = 0 # Initial value, will be filled by preparse_atoms_list()
        self.LSJ_list_indices = [] # Initial value, will be filled by preparse_atoms_list()
        self.LSJ_list_values = [] # Initial value, will be filled by preparse_atoms_list()
        self.param_list = []
        self.infilename = "model"+str(self.frag_num)+".xml"
        self.param_prefix = "m"+str(self.frag_num)+"_"

    def preparse_atoms_list(self, config_dom):
        '''Implements preparsing of XML-based description of atoms (centres with unpaired electrons)
        within the magnetochemical model. All "dirty" XML staff is confined here. Returns a list of dicts.'''
        atoms_dom = config_dom.getElementsByTagName('atoms')[0]
        atoms_list_dom = atoms_dom.getElementsByTagName('atom')
        proc_atoms = []
        j = 0
        for i in atoms_list_dom:
            j+=1
            atom = {}
            atom['element'] = i.getElementsByTagName('element')[0].childNodes[0].data
            number = i.getAttribute('number')
            if int(number)!=j :
                raise Exception('Atoms have to be numbered in order of description')
            atom['local_model'] = i.getElementsByTagName('local_model')[0].childNodes[0].data
            if atom['local_model'] == 'isotropic':
                pass
            elif atom['local_model'] == 'axial':
                if self.zeeman_symmetry == 'isotropic':
                    self.zeeman_symmetry = 'axial'
            elif atom['local_model'] == 'rhombic':
#                zeeman_symmetry = 'rhombic'
                raise Exception('Atom '+str(j)+': sorry, rhombic symmetry is not yet implemented')
            else: 
                raise Exception('Atom '+str(j)+': wrong local model requested')
            atom['spin_orbital'] = i.getElementsByTagName('spin_orbital')[0].childNodes[0].data
            if(atom['spin_orbital'] == 'LS'):
                atom['s'] = check_pos_one_half(i.getElementsByTagName('s')[0].childNodes[0].data)
                atom['l'] = check_pos_one_half(i.getElementsByTagName('l')[0].childNodes[0].data)
                self.LSJ_list_indices.append(number+'s')
                self.LSJ_list_values.append(atom['s'])
                if not fl_equal(atom['l'], 0.0):
                    self.LSJ_list_indices.append(number+'l')
                    self.LSJ_list_values.append(atom['l'])
                    self.total_l += int(atom['l'])
#               if atom['local_model'] != 'axial':
#               raise Exception(_('Atom ')+str(j)+_(': axial local model requested when L!=0'))
# No! We shall simply omit "delta*Lz^2"
            elif(atom['spin_orbital'] == 'JJ'):
                raise Exception('JJ model is not yet implemented, sorry')
            else:
                raise Exception('Atom '+str(j)+': wrong spin-orbital model requested')
            proc_atoms.append(atom)
        return proc_atoms

    def preparse_exchanges_list(self, config_dom):
        '''Implements preparsing of XML-based description of exchange interactions within the magnetochemical 
        model. All "dirty" XML staff is confined here. Returns a list of dicts.'''
        exchanges_dom = config_dom.getElementsByTagName('exchanges')[0]
        exchanges_list_dom = exchanges_dom.getElementsByTagName('exchange')
        proc_exs = []
        j = 0
        for i in exchanges_list_dom:
            j+=1
            ex = {}
            number = i.getAttribute('number')
            if int(number)!=j :
                raise Exception('Exchanges have to be numbered in order of description')
            ex['atom1'] = i.getElementsByTagName('atom1')[0].childNodes[0].data
            ex['atom2'] = i.getElementsByTagName('atom2')[0].childNodes[0].data
# TODO: Add check if these atoms exist
            ex['ex_model'] = i.getElementsByTagName('ex_model')[0].childNodes[0].data
# TODO: Add check of ex_model sanity
            proc_exs.append(ex)
        return proc_exs

    def preparse_simplifications_list(self, config_dom):
        '''Implements preparsing of XML-based description of parameters' simplifications 
        All "dirty" XML staff is confined here. Returns a substitution dictionary.'''
        simplifications_dom = config_dom.getElementsByTagName('simplifications')[0]
        simplifications_list_dom = simplifications_dom.getElementsByTagName('apply')
        subst_list = {}
        for i in simplifications_list_dom:
            s = i.childNodes[0].data.split('=')
            sub_arr = s[:-1]
            sub = s[-1:]
            for i in sub_arr:
                subst_list[self.param_prefix+i] = self.param_prefix+sub[0]
        return subst_list

    def build_ham_vectors_list(self):
        ham_vectors = [[]]
        for i in range(len(self.LSJ_list_indices)):
            key = self.LSJ_list_indices[i]
            value = self.LSJ_list_values[i]
            ham_vectors = [list_append2copy(i,j) for i in ham_vectors for j in projections_list(value)]
        return ham_vectors

    def gen_parameter(self, param):
        p_sub = param
        if self.subst_list.has_key(p_sub):
            p_sub = self.subst_list[p_sub]
        if p_sub not in self.param_list:
            self.param_list.append(p_sub)
        return p_sub

    def exchange(self):
        ex_summands = []
        for i in range(len(self.exs_list)):
            atom1 = self.exs_list[i]['atom1']
            atom2 = self.exs_list[i]['atom2']
            if self.exs_list[i]['ex_model'] == 'Heisenberg':
                J = self.gen_parameter(self.param_prefix+'Jz'+str(i+1))
                el1 = self.LSJ_list_indices.index(str(atom1)+"s") # Search for LSJ_list position corresponding to the moment, e.g. 1s
                el2 = self.LSJ_list_indices.index(str(atom2)+"s")
                ex_i = filter_zeros(const_by_matrix(-2, z1z2(self.ham_vectors,el1,el2)))
                ex_i_xy1 = filter_zeros(const_by_matrix(-1, plusminus(self.ham_vectors,el1,el2,self.LSJ_list_values[el1],self.LSJ_list_values[el2])))
                ex_i_xy2 = filter_zeros(const_by_matrix(-1, minusplus(self.ham_vectors,el1,el2,self.LSJ_list_values[el1],self.LSJ_list_values[el2])))
                ex_summands.append((J,merge_sum([ex_i,ex_i_xy1,ex_i_xy2])))
        return ex_summands

    def ZFS(self):
        ZFS_mat = []
        for i in range(len(self.atoms_list)):
            if self.atoms_list[i]['l']!=0:
                delta = self.gen_parameter(self.param_prefix+'delta'+str(i+1)) # Operator delta_Lz^2, part of LS operator
                el_l = self.LSJ_list_indices.index(str(i+1)+"l")
                ZFS_i = filter_zeros(z_square(self.ham_vectors, el_l))
                ZFS_mat.append((delta,ZFS_i))
            elif self.atoms_list[i]['local_model']=='axial' or self.atoms_list[i]['local_model']=='rhombic':
                D = self.gen_parameter(self.param_prefix+'D'+str(i+1)) # Operator delta_Lz^2, part of LS operator
                el = self.LSJ_list_indices.index(str(i+1)+"s")
                ZFS_i = filter_zeros(z_square(self.ham_vectors, el))
                ZFS_mat.append((D,ZFS_i))
        return ZFS_mat

    def LS(self):
        LS_mat = []
        for i in range(len(self.atoms_list)):
            if self.atoms_list[i]['l']!=0:
                _lambda = self.gen_parameter(self.param_prefix+'lambda'+str(i+1))
                eta = self.gen_parameter(self.param_prefix+'eta'+str(i+1))
                el = self.LSJ_list_indices.index(str(i+1)+"s")
                el_l = self.LSJ_list_indices.index(str(i+1)+"l")
                LS_i_z = filter_zeros(z1z2(self.ham_vectors,el,el_l))
                LS_i_xy_1 = filter_zeros(const_by_matrix(0.5, plusminus(self.ham_vectors,el,el_l,self.LSJ_list_values[el],self.LSJ_list_values[el_l])))
                LS_i_xy_2 = filter_zeros(const_by_matrix(0.5, minusplus(self.ham_vectors,el,el_l,self.LSJ_list_values[el],self.LSJ_list_values[el_l])))
                LS_mat.append((_lambda+'*'+eta, merge_sum([LS_i_z,LS_i_xy_2,LS_i_xy_2])))
        return LS_mat

    def z_zeeman(self):
        z_summands = []
        for i in range(len(self.atoms_list)):
            g_factor = self.gen_parameter(self.param_prefix+'gz'+str(i+1))
            el = self.LSJ_list_indices.index(str(i+1)+"s")
            z_zee_i = filter_zeros(z(self.ham_vectors,el))
            z_summands.append(('mu*Hz*'+g_factor, z_zee_i))
            if self.atoms_list[i]['l']!=0:
                eta = self.gen_parameter(self.param_prefix+'eta'+str(i+1))
                el_l = self.LSJ_list_indices.index(str(i+1)+"l")
                z_zee_i_l = filter_zeros(z(self.ham_vectors,el_l))
                z_summands.append(('mu*Hz*'+eta, z_zee_i_l))
        return z_summands

    def x_zeeman(self):
        x_summands = []
        for i in range(len(self.atoms_list)):
            g_factor = self.gen_parameter(self.param_prefix+'gx'+str(i+1))
            el = self.LSJ_list_indices.index(str(i+1)+"s")
            x_zee_i_1 = filter_zeros(const_by_matrix(0.5, plus(self.ham_vectors,el,self.LSJ_list_values[el])))
            x_zee_i_2 = filter_zeros(const_by_matrix(0.5, minus(self.ham_vectors,el,self.LSJ_list_values[el])))
            x_summands.append(('mu*Hx*'+g_factor, merge_sum([x_zee_i_1, x_zee_i_2])))
            L = self.atoms_list[i]['l']
            if L!=0:
                eta = self.gen_parameter(self.param_prefix+'eta'+str(i+1))
                el_l = self.LSJ_list_indices.index(str(i+1)+"l")
                x_zee_i_l_1 = filter_zeros(const_by_matrix(0.5, plus(self.ham_vectors,el_l,self.LSJ_list_values[el_l])))
                x_zee_i_l_2 = filter_zeros(const_by_matrix(0.5, minus(self.ham_vectors,el_l,self.LSJ_list_values[el_l])))
                x_summands.append(('mu*Hx*'+eta, merge_sum([x_zee_i_l_1,x_zee_i_l_2])))
        return x_summands

    def y_zeeman(self):
        y_summands = []
        for i in range(len(self.atoms_list)):
            g_factor = self.gen_parameter(self.param_prefix+'gy'+str(i+1))
            el = self.LSJ_list_indices.index(str(i+1)+"s")
            y_zee_i_1 = filter_zeros(const_by_matrix(complex('0.5j'), plus(self.ham_vectors,el,self.LSJ_list_values[el])))
            y_zee_i_2 = filter_zeros(const_by_matrix(complex('-0.5j'), minus(self.ham_vectors,el,self.LSJ_list_values[el])))
            y_summands.append(('mu*Hy*'+g_factor, merge_sum([y_zee_i_1, y_zee_i_2])))
            L = self.atoms_list[i]['l']
            if L!=0:
                eta = self.gen_parameter(self.param_prefix+'eta'+str(i+1))
                el_l = self.LSJ_list_indices.index(str(i+1)+"l")
                y_zee_i_l_1 = filter_zeros(const_by_matrix(complex('0.5j'), plus(self.ham_vectors,el_l,self.LSJ_list_values[el_l])))
                y_zee_i_l_2 = filter_zeros(const_by_matrix(complex('0.5j'), minus(self.ham_vectors,el_l,self.LSJ_list_values[el_l])))
                y_summands.append(('mu*Hy*'+eta, merge_sum([y_zee_i_l_1,y_zee_i_l_2])))
        return y_summands

    def parse(self, test=False):
        config_dom = xmlparser.parse(self.infilename) # XML parser function from xml.dom.minidom
        self.atoms_list = self.preparse_atoms_list(config_dom) # Atoms list
        self.exs_list = self.preparse_exchanges_list(config_dom) # Exchanges list
        self.subst_list = self.preparse_simplifications_list(config_dom)  # Simplifications list
        self.ham_vectors = self.build_ham_vectors_list()
        if not test:
            self.summands_common = []
            self.summands_common.extend(merge_by_coef(self.exchange()))
            if self.total_l!=0:
                self.summands_common.extend(merge_by_coef(self.LS()))
            if self.zeeman_symmetry!='isotropic':
                self.summands_common.extend(merge_by_coef(self.ZFS()))
                self.x_summands = self.summands_common[:]
                self.x_summands.extend(merge_by_coef(self.x_zeeman()))
                self.x_coefs = [item[0] for item in self.x_summands]
            self.z_summands = self.summands_common[:]
            self.z_summands.extend(merge_by_coef(self.z_zeeman()))
            self.z_coefs = [item[0] for item in self.z_summands]
    def read(self):
        self.zeeman_symmetry = open(self.param_prefix+"zeeman_symmetry","r").read().strip()
        self.total_l = int(open(self.param_prefix+"total_l","r").read().strip())
        self.param_list = [unicode(i.strip()) for i in open(self.param_prefix+"param_list","r").read().split(",")]
        self.z_coefs = [unicode(i.strip()) for i in open(self.param_prefix+"z_coefs","r").read().split(",")]
        if self.zeeman_symmetry!='isotropic':
            self.x_coefs = [unicode(i.strip()) for i in open(self.param_prefix+"x_coefs","r").read().split(",")]
            print self.x_coefs
        self.eigensolver = os.path.join(os.path.dirname(os.path.realpath(__file__)), mj_config_arr["eigensolver"])
        self.load_eigensolvers()
    def save_field(self, field, fname):
        f = open(self.param_prefix+fname, "w")
        f.write(field)
        f.close()
    def save(self):
        self.save_field(self.zeeman_symmetry,"zeeman_symmetry")
        self.save_field(str(self.total_l),"total_l")
        self.save_field(",".join(self.param_list),"param_list")
        self.save_field(",".join(self.z_coefs),"z_coefs")
        z_matrices_list_file = str(len(self.z_summands))+" "+str(len(self.ham_vectors))+os.linesep
        elem2str = lambda x:str(x[0]+1)+" "+str(x[1]+1)+" "+str(x[2]) # +1 refers to 0-starting -> 1-starting array convention change (for Fortran)
        for i in range(len(self.z_summands)):
            z_mtx_name = 'z_'+str(i)+'.mtx'
            self.save_field(str(len(self.z_summands[i][1]))+os.linesep+os.linesep.join(map(elem2str, self.z_summands[i][1])),z_mtx_name)
            z_matrices_list_file+=self.param_prefix+z_mtx_name+os.linesep
        self.save_field(z_matrices_list_file,'z_matrices_list')
        if self.zeeman_symmetry!='isotropic':
            self.save_field(",".join(self.x_coefs),"x_coefs")
            x_matrices_list_file = str(len(self.x_summands))+" "+str(len(self.ham_vectors))+os.linesep
            for i in range(len(self.x_summands)):
                x_mtx_name = 'x_'+str(i)+'.mtx'
                self.save_field(str(len(self.x_summands[i][1]))+os.linesep+os.linesep.join(map(elem2str, self.x_summands[i][1])),x_mtx_name)
                x_matrices_list_file+=self.param_prefix+x_mtx_name+os.linesep
            self.save_field(x_matrices_list_file,'x_matrices_list')
    def get_params_list(self):
        return self.param_list
    def load_eigensolvers(self):
        self.z_eigensolver = mj_eigensolver_caller(self.eigensolver, self.param_prefix+"z_matrices_list")
        if self.zeeman_symmetry != "isotropic":
             self.x_eigensolver = mj_eigensolver_caller(self.eigensolver, self.param_prefix+"x_matrices_list")
    def chi_z(self,temps,params_script):
        exec(params_script)
        exec("Hz=H")
        exec("Hx=H")
        z_coefs_values = []
        for i in range(len(self.z_coefs)):
            exec("val="+self.z_coefs[i])
            z_coefs_values.append(val)
        eig_H = self.z_eigensolver.calculate(z_coefs_values)

        dH = math.pow(10,-ddH)
        exec("H=H+dH")
        exec("Hz=H")
        exec("Hx=H")
        z_coefs_values = []
        for i in range(len(self.z_coefs)):
            exec("val="+self.z_coefs[i])
            z_coefs_values.append(val)
        eig_HdH = self.z_eigensolver.calculate(z_coefs_values)

        eig_shift = min(eig_H)
        eig_H_shifted = [i-eig_shift for i in eig_H]
        eig_HdH_shifted = [i-eig_shift for i in eig_HdH]
        print eig_H_shifted
        print eig_HdH_shifted
        x_arr = []
        for t in temps:
            n_sum = 0
            d_sum = 0
            for i in range(len(eig_H_shifted)):
                n_sum+=-((eig_HdH_shifted[i]-eig_H_shifted[i])/dH)*math.exp(-eig_H_shifted[i]/(k*t))
                d_sum+=math.exp(-eig_H_shifted[i]/(k*t))
            x_arr.append((t, (chi_const/H)*(n_sum/d_sum)))
        print x_arr
        return x_arr
    def chi_x(self,temps,params_script):
        exec(params_script)
        exec("Hz=H")
        exec("Hx=H")
        x_coefs_values = []
        for i in range(len(self.x_coefs)):
            exec("val="+self.x_coefs[i])
            x_coefs_values.append(val)
        eig_H = self.x_eigensolver.calculate(x_coefs_values)

        dH = math.pow(10,-ddH)
        exec("H=H+dH")
        exec("Hz=H")
        exec("Hx=H")
        x_coefs_values = []
        for i in range(len(self.x_coefs)):
            exec("val="+self.x_coefs[i])
            x_coefs_values.append(val)
        eig_HdH = self.x_eigensolver.calculate(x_coefs_values)

        eig_shift = min(eig_H)
        eig_H_shifted = [i-eig_shift for i in eig_H]
        eig_HdH_shifted = [i-eig_shift for i in eig_HdH]
        print eig_H_shifted
        print eig_HdH_shifted
        x_arr = []
        for t in temps:
            n_sum = 0
            d_sum = 0
            for i in range(len(eig_H_shifted)):
                n_sum+=-((eig_HdH_shifted[i]-eig_H_shifted[i])/dH)*math.exp(-eig_H_shifted[i]/(k*t))
                d_sum+=math.exp(-eig_H_shifted[i]/(k*t))
            x_arr.append((t, (chi_const/H)*(n_sum/d_sum)))
        print x_arr
        return x_arr

    def chi(self,temps,params_script):
        chi_z = self.chi_z(temps,params_script)
        if self.zeeman_symmetry != "isotropic":
            chi_z_coef = 1.0/3.0
            chi_x_coef = 1.0 - chi_z_coef
            chi_x = self.chi_x(temps,params_script)
            chi_total = [(temps[i],chi_z_coef*chi_z[i][1]+chi_x_coef*chi_x[i][1]) for i in range(len(chi_z))]
            return chi_total
        else:
            return chi_z

#a = mj_fragment_matrix(0)
#a.parse()
#a.save()
#a.read()
#a = mj_fragment_matrix(1)

#print projections_list(5.5)
